/********************************************************************
 *   rm - remove (delete) files                                     *
 *                                                                  *
 *   Syntax:     ex rm;'[-vfir] <filelist>'                         *
 *                                                                  *
 *      -v      verbose mode - report on files removed/failed       *
 *      -f      suppress warning messages if unable to remove files *
 *      -i      interactively remove files                          *
 *      -r      recurse down directory levels                       *
 ********************************************************************/

/*====================================================================
                            AMENDMENT HISTORY
                            ~~~~~~~~~~~~~~~~~
25/01/91  v1.0  DJW     Original version by D. J. Walker
01/04/91  v1.1  DJW     -f, -i, -r and -v flags added
06/06/91  v1.2  DJW     Version & copyright now displayed in usage 
                        screen
01/09/92  v1.4  DJW     Tidied up to remove all error and warnings
                        when compiled with c68 v3.04.
                        Output of message at end now suppressed if the verbose
                        flag is not set unless there are errors.
22 oct 92 v1.5  DJW     The verbose flag is forced to ON if the RM command
                        is executed directly from SuperBasic.
30 Aug 93 v1.55 DJW     Rebuilt to be QXL compatible
---------------------------------------------------------------------*/

char  _prog_name[] = "rm";
char  copyright[] = "(Remove files)  (c) copyright 1991 David J. Walker";
char  _version[] = "v1.55 ";

#include <qdos.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

/*
 *  Options to tailor program to QDOS environment
 */
long    _mneed = 4L * 1024L;
long    _stack = 4L * 1024L;
void    (*_consetup)() = consetup_title;    /* Console screen with Title bar */
int     (*_conwrite)() = NULL;              /* No Output translation needed */
int     (*_conread)() = NULL;               /* No Input translation needed */

int     errors = 0;
int     count = 0;
int     fflag = 0;          /* Force flag */
int     iflag = 0;          /* Interactive flag */
int     rflag = 0;          /* Recursion falg */
int     vflag = 0;          /* Verbose flag */
char    device[20];


/*============================================================== EXITS */
static
void    exits(int n)
/*      ~~~~~
 *  Exit.  If running directly under user control, give him a
 *  chance to see any messages before exiting.
 *---------------------------------------------------------------------*/
{
    if (n) {
        errno = n;
        perror ("Error");
    }
    exit (n);
}


/*============================================================== PROGRAM_ID */
static
void    Program_Id (void)
/*      ~~~~~~~~~~
 *--------------------------------------------------------------------------*/
{
    (void) fprintf (stderr,"%s %s %s\n\n",_prog_name, _version, copyright);
    return;
}


/*=================================================================== USAGE */
static
void    usage(char * str)
/*      ~~~~~
 *--------------------------------------------------------------------------*/
{
    Program_Id();
    (void) fprintf (stderr,"usage: rm [-fir] <filelist>\n");
    (void) fprintf (stderr,"       -f    suppress warnings and errors\n");
    (void) fprintf (stderr,"       -i    interactively remove files\n");
    (void) fprintf (stderr,"       -r    recursively remove directories\n");
    (void) fprintf (stderr,"       -v    verbose mode\n");
    (void) fprintf (stderr,"\n%s\n",str);
    exits(0);
}


/*================================================================= CONFIRM */
static
int     confirm (void)
/*      ~~~~~~~
 *--------------------------------------------------------------------------*/
{
    char t;

    (void) read( fileno(stdin), &t, (unsigned)1);
    (void) puts("");
    return (t == 'y' || t == 'Y');
}


/*============================================================ DELETE_FILES */
static
void    delete_files (char * name)
/*      ~~~~~~~~~~~~
 *  Delete the specified file(s).  If a sub-directory then use
 *  this routine recursively to descend a level.
 *
 *  Allowance must be made for the fact that there might be no
 *  device name present if the user is using Toolkit 2 default
 *  directories.
 *
 *  There also appears to be aproblem in deleting files with V2 disk
 *  drivers if the directory is open ?
 *--------------------------------------------------------------------------*/
{
    struct  qdirect direntry;
    long    chanid;
    char    ret_name[50];
    char    fullname[50];
    char    currentdir[50];
    char    devicename[20];
    int     devlen, dirlen, extra;
    int     devtype = DIRDEV;
    long    dirpos;

    if (isdevice(name,&devtype)) {
        devlen=0;
        do {
            devicename[devlen] = name[devlen];
        } while (name[devlen++] != '_'); 
        devicename[devlen] = '\0';           /* NULL: terminate */
    } else {
        (void) strcpy(devicename,"");
    }
    devlen = strlen(devicename);
    if ((chanid = open_qdir(name)) < 0 ) {
        if (!fflag) {
            (void) fprintf (stderr, "rm: can't open directory '%s'\n", name);
        }
        errors++;
        return;
    }
    if (rflag) {
        while (read_qdir(chanid,name,ret_name,&direntry,4)==1) {
            (void) strcpy (fullname,devicename);
            (void) memcpy (&fullname[devlen],direntry.d_name,(unsigned)direntry.d_szname);
            fullname[direntry.d_szname + devlen]='\0';
            dirpos = fs_pos (chanid, 0L, 1);    /* save position */
            (void) io_close (chanid);           /* ... and close */
            delete_files(fullname);             /* descend a level */
            chanid = open_qdir(name);           /* re-open directory */
            (void) fs_pos (chanid, dirpos, 0);  /* ... and reset position */
        }
    }
    /*
     * Get the current data directory, 
     * removing any device name
     */
    if (getcwd (currentdir,(int)sizeof(currentdir))) {
        if (isdevice (currentdir, &extra)) {
            char *ptr = strpbrk(currentdir,"_");
            if (ptr)
                (void) strcpy(currentdir,++ptr);
        }
    } else {
        (void) strcpy(currentdir,"");
    }
    dirlen = strlen (currentdir);
    /*
     * Read through directory looking for files
     * that match the selection criteria
     */
    (void) fs_pos (chanid, 0L, 0);         /* reset to start */
    while (read_qdir(chanid,name,ret_name,&direntry,0)==1) {
        count++;
        (void) strcpy (fullname,devicename);
        (void) memcpy (&fullname[devlen],direntry.d_name,(size_t)direntry.d_szname);
        fullname[direntry.d_szname + devlen]='\0';
        if (!strnicmp(currentdir,fullname,(size_t)dirlen)) {
            (void) strcpy (fullname, &fullname[dirlen]);
        }
        if (iflag) {
            (void) fprintf (stderr, "rm: remove %s? (y/n) ", fullname);
            if (!confirm())  continue;
        }
        dirpos = fs_pos (chanid, 0L, 1);    /* save position */
        (void) io_close (chanid);           /* ... and close */
        if (unlink(fullname)) {
            if (!fflag) {
                (void) poserr (fullname);
            }
            errors++;
        } else {
            if (vflag)
                (void) fprintf (stderr, "rm: %s removed\n", fullname);
        }
        chanid = open_qdir(name);           /* re-open directory */
        (void) fs_pos (chanid, dirpos, 0);  /* ... and reset position */
    } /* End of while loop */
    (void)io_close (chanid);
}



/*==================================================================== MAIN */
extern
void    main (int argc, char * argv[])
/*      ~~~~
 *--------------------------------------------------------------------------*/
{
    char *opt;
    int exstatus;

    if (iscon(fgetchid(stdout),0) && !isnoclose(fileno(stdout)))
        vflag++;
    if (argc < 2) 
        usage("incorrect number of parameters");
    ++argv;
    --argc;
    while (**argv == '-') {
        opt = *argv;
        while (*++opt != '\0') {
            switch (*opt) {
                case 'f':
                case 'F':
                        fflag++;    break;
                case 'i':
                case 'I':
                        iflag++;    break;
                case 'r':
                case 'R':
                        rflag++;    break;
                case 'v':
                case 'V':
                        Program_Id ();
                        vflag++;    break;
                default:
                        usage("unknown option");
            } /* end of switch */
        }
        argc--;
        ++argv;
    }
    if (argc < 1) {
        usage("no file names specified");
    }
    while (argc--) {
        delete_files(*argv++);
    }
    exstatus = (errors == 0 ? 0 : -1);
    if (fflag) {
        exstatus = 0;
    }
    if ( ! vflag && exstatus == 0) {
        _endmsg = NULL;
    }
    exits(exstatus);
}

